<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

namespace enrol_ubxsshib\form;

defined('MOODLE_INTERNAL') || die();

use moodleform;
use course_enrolment_manager;
use enrol_ubxsshib\ubxsshib_enrol_method;

require_once($CFG->libdir . '/formslib.php');
require_once($CFG->dirroot . '/enrol/locallib.php');
require_once($CFG->dirroot . '/enrol/ubxsshib/locallib.php');

/**
 * Adds new instance of enrol_ubxsshib to specified course
 * or edits current instance.
 *
 * @package    enrol_ubxsshib
 * @copyright  2011 - MAPI Université de Bordeaux
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */
class edit_form extends moodleform {

    /**
     * {@inheritdoc}
     *
     * @return void
     * @throws \coding_exception
     * @throws \dml_exception
     */
    protected function definition() {
        global $PAGE, $CFG, $DB;

        $mform = $this->_form;
        list($instance, $plugin, $context) = $this->_customdata;
        $courseid = $instance->courseid;

        if (!$instance->id) {
            $instance = (object)array_merge((array)$instance, $plugin->get_instance_defaults());
        }

        $mform->addElement('header', 'header', get_string('pluginname', 'enrol_ubxsshib'));

        $namefield = $mform->addElement('text', 'name', get_string('custominstancename', 'enrol'));
        $mform->setType('name', PARAM_TEXT);
        $mform->addRule('name', null, 'required', null, 'client');
        $namefield->updateAttributes(['data-apo-field' => 'name']);

        $options = [
            ENROL_INSTANCE_ENABLED => get_string('yes'),
            ENROL_INSTANCE_DISABLED => get_string('no'),
        ];
        $mform->addElement('select', 'status', get_string('status', 'enrol_ubxsshib'), $options);
        $mform->setDefault('status', $plugin->get_config('status'));

        $options = [
            ubxsshib_enrol_method::CODE_ETAPE => get_string('etape_choice', 'enrol_ubxsshib'),
            ubxsshib_enrol_method::CODE_ELEMENTPEDAGO => get_string('elp_choice', 'enrol_ubxsshib'),
            ubxsshib_enrol_method::CODE_EPREUVE => get_string('exam_choice', 'enrol_ubxsshib'),
        ];
        $typefield = $mform->addElement('select', 'customchar1', 'Type', $options);
        $mform->setDefault('customchar1', ubxsshib_enrol_method::CODE_ETAPE);
        $typefield->updateAttributes(['data-apo-field' => 'codetype']);

        $codefield = $mform->createElement('text', 'customchar2', get_string('apogeeid', 'enrol_ubxsshib'));
        $mform->setType('customchar2', PARAM_TEXT);
        $codefield->updateAttributes(['data-apo-field' => 'codevalue']);

        $searchbutton = $mform->createElement('button', 'searchbutton', get_string('searchbutton', 'enrol_ubxsshib'));

        $codearray = [];
        $codearray[] =& $codefield;
        $codearray[] =& $searchbutton;
        $mform->addGroup($codearray, 'codearray', get_string('apogeeid', 'enrol_ubxsshib'), [' '], false);
        $mform->addRule('codearray', null, 'required', null, 'client');
        $mform->addHelpButton('codearray', 'apogeeid', 'enrol_ubxsshib');

        $manager = new course_enrolment_manager($PAGE, $DB->get_record('course', ['id' => $courseid], '*', MUST_EXIST));
        $roles = array_reverse($manager->get_assignable_roles(false), true);
        $mform->addElement('select', 'roleid', get_string('assignrole', 'enrol_ubxsshib'), $roles);
        $mform->setDefault('roleid', $plugin->get_config('roleid'));
        $mform->addHelpButton('roleid', 'assignrole', 'enrol_ubxsshib');

        $mform->addElement('date_selector', 'enrolstartdate', get_string('enrolstartdate', 'enrol_self'), ['optional' => true]);
        $mform->addHelpButton('enrolstartdate', 'enrolstartdate', 'enrol_ubxsshib');

        $mform->addElement('date_selector', 'enrolenddate', get_string('enrolenddate', 'enrol_self'), ['optional' => true]);
        $mform->addHelpButton('enrolenddate', 'enrolenddate', 'enrol_ubxsshib');

        $mform->addElement('hidden', 'id');
        $mform->setType('id', PARAM_INT);
        $mform->addElement('hidden', 'courseid');
        $mform->setType('courseid', PARAM_INT);

        $groups = [];
        $strnogroup = get_string('nogroup', 'group');
        $groups[ubxsshib_enrol_method::NO_GROUP] = (object)[
            'id' => ubxsshib_enrol_method::NO_GROUP,
            'courseid' => $courseid,
            'idnumber' => '',
            'name' => $strnogroup,
            'description' => '',
            'descriptionformat' => FORMAT_HTML,
            'enrolmentkey' => '',
            'picture' => 0,
            'hidepicture' => 0,
            'timecreated' => 0,
            'timemodified' => 0,
        ];
        $groups = array_merge($groups, $DB->get_records('groups', ['courseid' => $courseid], 'name'));
        $options = [];
        foreach ($groups as $group) {
            $options[$group->id] = strip_tags(format_string($group->name));
        }
        $mform->addElement('header', 'headeroptions', 'Options');
        $mform->addElement('select', 'customint1', get_string('enrolgroup', 'enrol_ubxsshib'), $options);
        $mform->setDefault('customint1', ubxsshib_enrol_method::NO_GROUP);
        $mform->addHelpButton('customint1', 'enrolgroup', 'enrol_ubxsshib');

        $this->add_action_buttons(true, ($instance->id ? null : get_string('addinstance', 'enrol')));

        $this->set_data($instance);

        $PAGE->requires->js_call_amd('enrol_ubxsshib/search', 'initEditForm');
    }

    /**
     * {@inheritdoc}
     *
     * @return void
     * @throws \dml_exception
     */
    public function definition_after_data() {
        global $DB;

        $mform = $this->_form;
        $id = intval($mform->getElementValue('id'));

        if (0 !== $id) {
            $nbenrolled = $DB->count_records('user_enrolments', ['enrolid' => $id]);
            if (0 < $nbenrolled) {
                $mform->freeze('customint1');
                $mform->freeze('roleid');
            }
        }
    }

    /**
     * {@inheritdoc}
     *
     * @param array $data array of ("fieldname"=>value) of submitted data
     * @param array $files array of uploaded files "element_name"=>tmp_file_path
     *
     * @return array of "element_name"=>"error_description" if there are errors,
     *          or an empty array if everything is OK (true allowed for backwards compatibility too).
     */
    public function validation($data, $files) {
        global $DB, $CFG;
        $errors = parent::validation($data, $files);

        if (empty($data['customchar2'])) {
            $errors['codeapogee'] = 'Veuillez saisir un code Apogee.';
        }

        if (!empty($data['customint1']) && $data['customint1'] != ubxsshib_enrol_method::NO_GROUP) {
            if (!enrol_ubxsshib_check_coursegroup_exists($data['customint1'])) {
                $errors['group'] = 'Le groupe n\'existe pas ou plus.';
            }
        }

        return $errors;
    }
}
