<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

/**
 * Class to intercat with LDAP server.
 *
 * @package    enrol_ubxsshib
 * @copyright  2020 - MAPI Université de Bordeaux
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

namespace enrol_ubxsshib;

defined('MOODLE_INTERNAL') || die();

use enrol_ubxsshib\user_info;
use enrol_ubxsshib\group_info;
use enrol_ubxsshib\ubxsshib_enrol_method;
use LDAP\Connection;

require_once($CFG->dirroot . '/enrol/ubxsshib/classes/IGroupProvider.php');
require_once($CFG->dirroot . '/enrol/locallib.php');

/**
 * Provides users info from LDAP source
 */
class ldap_provider implements IGroupProvider {
    /**
     * @var Connection
     */
    protected $ldapconnection;

    /**
     * @var \enrol_plugin|null
     */
    protected $plugin;

    /**
     * ldap_provider constructor.
     */
    public function __construct() {
        $this->plugin = enrol_get_plugin('ubxsshib');
    }

    /**
     * Closing ldap connection on destruction.
     *
     */
    public function __destruct() {
        if ($this->ldapconnection != null) {
            ldap_close($this->ldapconnection);
            $this->ldapconnection = null;
        }
    }

    /**
     * Returns a connection to ldap server.
     *
     * @return mixed connection result or false.
     */
    private function get_connection() {
        global $CFG;

        if ($this->ldapconnection == null) {
            require_once($CFG->libdir . '/ldaplib.php');

            $plugin = $this->plugin;
            $ldapurl = $plugin->get_config('host_url');
            $ldapversion = $plugin->get_config('ldap_version');
            $ldapbinddn = $plugin->get_config('bind_dn');
            $ldapbindpw = $plugin->get_config('bind_pw');
            $ldapoptderef = $plugin->get_config('opt_deref');
            $ldapusertype = 'default';
            $debuginfo = '';
            $ldapstarttls = $plugin->get_config('start_tls');

            $ldapconn = ldap_connect_moodle($ldapurl, $ldapversion,
                $ldapusertype, $ldapbinddn,
                $ldapbindpw, $ldapoptderef,
                $debuginfo, $ldapstarttls);
            if (false === $ldapconn) {
                $this->ldapconnection = null;
                debugging("enrol_ubxsshib:enrol_ubxsshib_ldapconnection - Cannot connect to ldap server");
            } else {
                $this->ldapconnection = $ldapconn;
            }
        }

        return $this->ldapconnection;
    }

    /**
     * Returns array of Ldap entries with $attributes that match $filter criterion.
     * @param string $base
     * @param string $filter
     * @param string[] $attributes
     * @return array
     */
    private function retrieve($base, $filter, $attributes) {
        $ldapconn = $this->get_connection();
        $entries = [];
        if ($ldapconn) {
            if ($search = ldap_search($ldapconn, $base, $filter, $attributes)) {
                $entries = ldap_get_entries($ldapconn, $search);
            }
        }

        return $entries;
    }

    /**
     * Returns userinfo hydrated with provided Ldap entry.
     * @param array $ldapentry
     * @return user_info
     */
    private function get_userinfo_from_ldapentry($ldapentry, $getroles = false) {
        $plugin = $this->plugin;
        $ldapusernameproperty = strtolower($plugin->get_config('username_property'));
        $ldapmemberof = strtolower($plugin->get_config('memberof_property'));
        $username = $ldapentry[$ldapusernameproperty][0];

        $cache = \cache::make('enrol_ubxsshib', 'userinfo');
        $userinfo = $cache->get($username);

        if (false === $userinfo) {
            $userinfo = new user_info();
            $userinfo->id = 0;
            $userinfo->hasrolesset = false;
            $userinfo->username = $ldapentry[$ldapusernameproperty][0];
            $userinfo->lastname = isset($ldapentry['sn']) ? $ldapentry['sn'][0] : '';
            $userinfo->firstname = isset($ldapentry['givenname']) ? $ldapentry['givenname'][0] : '';
            $userinfo->email = isset($ldapentry['mail']) ? $ldapentry['mail'][0] : '';
            $userinfo->status = isset($ldapentry['ubxstatutcompte']) ? $ldapentry['ubxstatutcompte'][0] : '';
            $groups = $this->extract_groups($ldapentry[$ldapmemberof] ?? []);
            $userinfo->etapes = $groups[ubxsshib_enrol_method::CODE_ETAPE];
            $userinfo->elements = $groups[ubxsshib_enrol_method::CODE_ELEMENTPEDAGO];
            $userinfo->epreuves = $groups[ubxsshib_enrol_method::CODE_EPREUVE];

            $cache->set($username, $userinfo);
        }

        if ($getroles && false === $userinfo->hasrolesset) {
            $userinfo->mustbesystemcoursecreator = $this->is_systemcoursecreator($userinfo);
            $userinfo->mustbesandboxcoursecreator = $this->is_sandboxcoursecreator($userinfo);
            $userinfo->hasrolesset = true;

            $cache->set($username, $userinfo);
        }

        return $userinfo;
    }

    /**
     * Returns Ldap groups of type $codetype containing $searchtext in description.
     * @param string $searchtext
     * @param string $grouptype
     * @return enrol_ubxsshib_groupinfo[]
     */
    public function search_groups($searchtext, $grouptype) {
        $plugin = $this->plugin;

        $ldapgroupdnbase = $plugin->get_config('group_dn_base');

        switch ($grouptype) {
            case ubxsshib_enrol_method::CODE_ETAPE:
                $ldapgroupnameproperty = $plugin->get_config('group_etape_name_property');
                $ldapgrouppatterns = preg_split('/\r\n|\r|\n/', $plugin->get_config('group_etape_pattern'), -1, PREG_SPLIT_NO_EMPTY);
                $ldapgroupsearchproperty = $plugin->get_config('group_etape_search_property');
                break;
            case ubxsshib_enrol_method::CODE_ELEMENTPEDAGO:
                $ldapgroupnameproperty = $plugin->get_config('group_elp_name_property');
                $ldapgrouppatterns = preg_split('/\r\n|\r|\n/', $plugin->get_config('group_elp_pattern'), -1, PREG_SPLIT_NO_EMPTY);
                $ldapgroupsearchproperty = $plugin->get_config('group_elp_search_property');
                break;
            case ubxsshib_enrol_method::CODE_EPREUVE:
                $ldapgroupnameproperty = $plugin->get_config('group_exam_name_property');
                $ldapgrouppatterns = preg_split('/\r\n|\r|\n/', $plugin->get_config('group_exam_pattern'), -1, PREG_SPLIT_NO_EMPTY);
                $ldapgroupsearchproperty = $plugin->get_config('group_exam_search_property');
                break;
        }

        // The '*' characters that are present in Apogee codes are substituted with '_-_' in LDAP.
        $searchtext = str_replace('*', '_-_', $searchtext);
        $texts = preg_split('/\s+/', $searchtext, -1, PREG_SPLIT_NO_EMPTY);
        $nbtexts = count($texts);
        $textfilter = implode(
            '',
            array_map(function ($t) use ($ldapgroupsearchproperty) {
                return "($ldapgroupsearchproperty=*$t*)";
            },
                $texts
            )
        );

        $results = [];
        foreach ($ldapgrouppatterns as $pattern) {
            $ldapgrouppattern = str_replace('%CODE%', '*', $pattern);
            $grouptypefilter = "($ldapgroupnameproperty=$ldapgrouppattern)";

            if ($nbtexts == 1) {
                // Looking for exact text in code value or wilcarded in description.
                $codefilter = "($ldapgroupnameproperty=" . str_replace('%CODE%', $texts[0], $pattern) . ")";
                $filter = "(|$codefilter(&$grouptypefilter$textfilter))";
            } else {
                // Looking for description containing all texts.
                $filter = "(&$grouptypefilter$textfilter)";
            }
            $attributes = [$ldapgroupnameproperty, $ldapgroupsearchproperty];
            $entries = $this->retrieve($ldapgroupdnbase, $filter, $attributes);
            $nbentries = $entries['count'];
            for ($i = 0; $i < $nbentries; $i++) {
                $entry = $entries[$i];
                $matches = [];
                $ldapgroup = $entry[$ldapgroupnameproperty][0];
                if (preg_match('/^' . str_replace('%CODE%', '(?P<code>.*)', str_replace('*', '(?:.*)', $pattern)) . '/', $ldapgroup, $matches) == 1) {
                    $result = new group_info();
                    // Restoring substituted '*' in Apogee codes.
                    $result->code = str_replace('_-_', '*', strtoupper($matches['code']));
                    $result->type = $grouptype;
                    $result->id = $ldapgroup;
                    $result->description = $entry[$ldapgroupsearchproperty][0];
                    $results[] = $result;
                }
            }
        }

        usort($results, function ($ga, $gb) {
            $desccomp = strcasecmp($ga->description, $gb->description);
            return $desccomp == 0 ?
                strcasecmp($ga->code, $gb->code) :
                $desccomp;
        });

        return $results;
    }

    /**
     * Returns group of type $codetype identified by its $groupecode.
     * If $withmembers is true, members are also extracted.
     * @param string $grouptype
     * @param string $groupcode
     * @param boolean $withmembers
     * @return enrol_ubxsshib_groupinfo
     */
    public function get_group($grouptype, $groupcode, $withmembers = false) {
        $plugin = $this->plugin;
        $ldapgroupdnbase = $plugin->get_config('group_dn_base');

        $grp = new group_info();
        $grp->type = $grouptype;
        $grp->code = $groupcode;

        switch ($grouptype) {
            case ubxsshib_enrol_method::CODE_ETAPE:
                $ldapgroupnameproperty = $plugin->get_config('group_etape_name_property');
                $ldapgrouppatterns = $plugin->get_config('group_etape_pattern');
                $ldapgroupsearchproperty = $plugin->get_config('group_etape_search_property');
                $pattern = str_replace('%CODE%', $groupcode, $ldapgrouppattern);
                $filter = "$ldapgroupnameproperty=" . $pattern;
                break;
            case ubxsshib_enrol_method::CODE_ELEMENTPEDAGO:
                $ldapgroupnameproperty = $plugin->get_config('group_elp_name_property');
                $ldapgrouppatterns = $plugin->get_config('group_elp_pattern');
                $ldapgroupsearchproperty = $plugin->get_config('group_elp_search_property');

                break;
            case ubxsshib_enrol_method::CODE_EPREUVE:
                $ldapgroupnameproperty = $plugin->get_config('group_exam_name_property');
                $ldapgrouppatterns = $plugin->get_config('group_exam_pattern');
                $ldapgroupsearchproperty = $plugin->get_config('group_exam_search_property');
                $groupcode = str_replace('*', '_-_', $groupcode);

                $pattern = str_replace('%CODE%', $groupcode, $ldapgrouppattern);
                $filter = "$ldapgroupnameproperty=" . $pattern;
                break;
        }
        $filter = '';
        $lines = preg_split('/\r\n|\r|\n/', $ldapgrouppatterns, -1, PREG_SPLIT_NO_EMPTY);
        foreach ($lines as $line) {
            $ldapgrouppattern = str_replace('%CODE%', $groupcode, $line);
            $filter = "$filter($ldapgroupnameproperty=$ldapgrouppattern)";
        }
        if (count($lines) > 1) {
            $filter = "(|$filter)";
        }

        $attributes = ['member', $ldapgroupsearchproperty, $ldapgroupnameproperty];
        $entries = $this->retrieve($ldapgroupdnbase, $filter, $attributes);
        if (isset($entries) && $entries['count'] > 0) {
            $entry = $entries[0];
            $grp->description = $entry[$ldapgroupsearchproperty][0];
            if ($withmembers) {
                $members = $this->search_users(['memberOf' => $ldapgroupnameproperty. '=' . $entry[$ldapgroupnameproperty][0]]);
                $keyedmembers = [];
                foreach ($members as $member) {
                    $keyedmembers[$member->username] = $member;
                }
                $grp->members = $keyedmembers;
            }
        } else {
            $grp = null;
        }

        return $grp;
    }

    /**
     * Returns Ldap users matching $criteria.
     * @param string[] $criteria
     * @return user_info[]
     */
    public function search_users($criteria) {
        $username = array_key_exists('username', $criteria) ? $criteria['username'] : '';
        $lastname = array_key_exists('lastname', $criteria) ? $criteria['lastname'] : '';
        $memberof = array_key_exists('memberOf', $criteria) ? $criteria['memberOf'] : '';
        $plugin = $this->plugin;
        $ldapusernameproperty = $plugin->get_config('username_property');
        $ldapuserdnbase = $plugin->get_config('user_dn_base');
        $ldapmemberofproperty = $plugin->get_config('memberof_property');
        $ldaplastnameproperty = 'sn';
        $ldapgroupdnbase = $plugin->get_config('group_dn_base');

        $filteruid = (isset($username) && '' !== $username ? "($ldapusernameproperty=$username)" : "($ldapusernameproperty=*)");
        $filterlastname = (isset($lastname) && '' !== $lastname ? "($ldaplastnameproperty=*$lastname*)" : '');
        $filtermembership = (isset($memberof) && '' != $memberof ? "($ldapmemberofproperty=$memberof,$ldapgroupdnbase)" : '');
        $filter = "(&$filteruid$filterlastname$filtermembership)";
        $attributes = [$ldapusernameproperty, 'sn', 'givenname', 'mail', 'ubxstatutcompte', $ldapmemberofproperty];
        $entries = array_change_key_case($this->retrieve($ldapuserdnbase, $filter, $attributes), CASE_LOWER);
        $results = [];
        $nbentries = (isset($entries) && $entries['count'] > 0 ? $entries['count'] : 0);
        for ($i = 0; $i < $nbentries; $i++) {
            $entry = $entries[$i];
            $result = $this->get_userinfo_from_ldapentry($entry, false);
            $results[] = $result;
        }

        return $results;
    }

    /**
     * Returns userinfo for the provided $username.
     * @param string $username
     * @return user_info
     */
    public function get_userinfo($username, $withroles = false) {
        $plugin = $this->plugin;

        $result = new user_info();

        $ldapuserdnbase = $plugin->get_config('user_dn_base');
        $ldapusernameproperty = $plugin->get_config('username_property');
        $ldapmemberof = $plugin->get_config('memberof_property');
        $attributes = [$ldapusernameproperty, 'sn', 'givenname', 'mail', 'ubxstatutcompte', $ldapmemberof];
        $filter = "($ldapusernameproperty=$username)";
        $entries = $this->retrieve($ldapuserdnbase, $filter, $attributes);

        if ($entries && $entries['count'] > 0) {
            $entry = $entries[0];
            $result = $this->get_userinfo_from_ldapentry($entry, $withroles);
        } else {
            $result = null;
        }

        return $result;
    }

    /**
     * Indicates if user's affiliation allows user to be system creator.
     *
     * @param array $affiliation Primary affiliation of the user
     * @return bool
     */
    protected function is_systemcoursecreator(user_info $userinfo) {
        $hascreationrights = false;

        $ldapconn = $this->get_connection();
        if ($ldapconn) {
            $plugin = enrol_get_plugin('ubxsshib');
            $ldapusername = $plugin->get_config('username_property');
            $ldapuserdnbase = $plugin->get_config('user_dn_base');
            $attributes = [$ldapusername];
            $ldapfilter = trim($plugin->get_config('systemcoursecreator_filter'));
            if ($ldapfilter != '') {
                $username = $userinfo->username;
                $filter = "(&($ldapusername=$username)$ldapfilter)";
                $search = ldap_search($ldapconn, $ldapuserdnbase, $filter, $attributes);
                $entries = ldap_get_entries($ldapconn, $search);
                $hascreationrights = ($entries['count'] > 0);
            }
        }

        return $hascreationrights;
    }


    /**
     * Indicates if user's affiliation allows user to be course creator in the sandbox.
     *
     * @param array $affiliation Primary affiliation of the user
     * @return bool
     */
    protected function is_sandboxcoursecreator(user_info $userinfo) {
        $hascreationrights = false;

        $ldapconn = $this->get_connection();
        if ($ldapconn) {
            $plugin = enrol_get_plugin('ubxsshib');
            $ldapusername = $plugin->get_config('username_property');
            $ldapuserdnbase = $plugin->get_config('user_dn_base');
            $attributes = [$ldapusername];
            $ldapfilter = trim($plugin->get_config('sandboxcoursecreator_filter'));

            if ($ldapfilter != '') {
                $username = $userinfo->username;
                $filter = "(&($ldapusername=$username)$ldapfilter)";
                $search = ldap_search($ldapconn, $ldapuserdnbase, $filter, $attributes);
                $entries = ldap_get_entries($ldapconn, $search);
                $hascreationrights = ($entries['count'] > 0);
            }
        }

        return $hascreationrights;
    }

    /**
     * Returning all group codes present in the LDAP groups,
     * grouped by group type.
     *
     * @param array $ldapgroups
     * @param $ldapgroups
     *
     * @return array[]
     */
    protected function extract_groups($ldapgroups) {
        $groups = [
            ubxsshib_enrol_method::CODE_ETAPE => [],
            ubxsshib_enrol_method::CODE_ELEMENTPEDAGO => [],
            ubxsshib_enrol_method::CODE_EPREUVE => [],
        ];

        if (!isset($ldapgroups) || (isset($ldapgroups) && isset($ldapgroups['count']) && $ldapgroups['count'] == 0)) {
            return $groups;
        }

        $plugin = $this->plugin;
        $ldapgroupdnbase = $plugin->get_config('group_dn_base');
        $filters = [];

        // Creation of all regex filter strings for each group type.
        $config = [
            ubxsshib_enrol_method::CODE_ETAPE => ['nameproperty' => 'group_etape_name_property', 'grouppattern' => 'group_etape_pattern'],
            ubxsshib_enrol_method::CODE_ELEMENTPEDAGO => ['nameproperty' => 'group_elp_name_property', 'grouppattern' => 'group_elp_pattern'],
            ubxsshib_enrol_method::CODE_EPREUVE => ['nameproperty' => 'group_exam_name_property', 'grouppattern' => 'group_exam_pattern'],
        ];
        foreach ($config as $type => $typeconfig) {
            $ldapgroupnameproperty = $plugin->get_config($typeconfig['nameproperty']);
            $patterns = preg_split('/\r\n|\r|\n/', $plugin->get_config($typeconfig['grouppattern']), -1, PREG_SPLIT_NO_EMPTY);
            foreach ($patterns as $pattern) {
                $ldapgrouppattern = str_replace(
                    '%CODE%',
                    '(?P<codevalue>.*)',
                    str_replace('*', '(?:.*)', $pattern)
                );
                $filter = '/^' . $ldapgroupnameproperty . '=' . $ldapgrouppattern . ',' . $ldapgroupdnbase . '/i';
                $filters[$filter] = $type;
            }
        }

        // Iterating on each group to find matching pattern in order to define its correct type.
        foreach ($ldapgroups as $grp) {
            foreach ($filters as $filter => $type) {
                $matches = [];
                if (preg_match($filter, $grp, $matches) == 1) {
                    $code = strtoupper($matches['codevalue']);
                    $code = str_replace('_-_', '*', $code);
                    array_push($groups[$type], $code);
                    break;
                }
            }
        }

        return $groups;
    }
}
