<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

/**
 * Force enrolments task
 * @package    enrol_ubxsshib
 * @copyright  2020 MAPI Université de Bordeaux
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

namespace enrol_ubxsshib\task;

defined('MOODLE_INTERNAL') || die();

use context_course;
use text_progress_trace;
use enrol_ubxsshib\ubxsshib_enrol_method;
use enrol_ubxsshib\event\forceenrol_started;
use enrol_ubxsshib\event\forceenrol_completed;

class force_enrolments extends \core\task\adhoc_task {

    /**
     * Run task for synchronizing users.
     */
    public function execute() {
        global $DB;

        /** @var \enrol_ubxsshib\ubxsshib_enrol_method $method */
        $method = $this->get_method();

        /** @var enrol_ubxsshib_plugin $plugin */
        $plugin = enrol_get_plugin('ubxsshib');
        $trace = new text_progress_trace();

        // Check if enrol method still exists.
        $methodid = $this->get_custom_data()->methodid;
        if (false === $method) {
            $trace->output("Enrol method #$methodid was not found => Exiting");
            return;
        }

        if ($plugin->get_config('inhibit_all', false)) {
            $trace->output('All methods are disabled in plugin settings.');
            return;
        }

        /** @var \user $owner */
        $owner = $this->get_owner();
        $eventparams = [
            'context' => $context = context_course::instance($method->enrol_instance()->courseid),
            'other' => [
                'methodid' => $method->get_id(),
                'methodname' => $method->enrol_instance()->name,
                'userid' => $owner->id,
            ],
        ];

        $startevent = forceenrol_started::create($eventparams);
        $startevent->trigger();

        $trace->output('Synchronization of method "' . $method->enrol_instance()->name . '" in course #' . $method->enrol_instance()->courseid . ' launched by user #' . $owner->id . ' (' . $owner->firstname . ' ' . $owner->lastname . ')');
        try {
            $plugin->sync_enrolments($trace, $method);
        } catch (Exception $e) {
            $trace->output('An error occured while method was being synchronized.');
            return;
        }

        $endevent = forceenrol_completed::create($eventparams);
        $endevent->trigger();

        $trace->output('Method has been successfully synchronized.');
        $trace->finished();
    }

    /**
     * Return ubxsshib_enrol_method or false if it does not
     * exists anymore.
     * @global type $DB
     * @return mixed
     */
    public function get_method() {
        global $DB;

        $data = $this->get_custom_data();
        $methodid = $data->methodid;
        $enrol = $DB->get_record('enrol', ['id' => $methodid], '*', IGNORE_MISSING);

        if (!$enrol) {
            $method = false;
        } else {
            $method = ubxsshib_enrol_method::get($enrol);
        }

        return $method;
    }

    public function get_owner() {
        global $DB;

        $owner = $DB->get_record('user', ['id' => $this->get_userid()], '*', MUST_EXIST);
        return $owner;
    }
}
