<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

/**
 * Class to manage ubxsshib enrolment method instances.
 *
 * @package    enrol_ubxsshib
 * @copyright  2018 - MAPI Université de Bordeaux
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

namespace enrol_ubxsshib;

defined('MOODLE_INTERNAL') || die();

require_once($CFG->dirroot . '/cohort/lib.php');

/**
 * Class to manage ubxsshib enrolment method instances.
 *
 * @package    enrol_ubxsshib
 * @copyright  2018 - MAPI Université de Bordeaux
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */
class ubxsshib_cohort {
    const SYNCMODE_SYNC = 'sync';
    const SYNCMODE_ADD = 'add';

    private $instance = null;
    private $plugin = null;
    private $pattern = "/ubgroup:(" . self::SYNCMODE_SYNC . "|" . self::SYNCMODE_ADD . "):(.*)/";

    public function __construct(\stdClass $cohortinstance) {
        $this->instance = $cohortinstance;
        $this->plugin = enrol_get_plugin('ubxsshib');
    }

    public function get_id() {
        return $this->instance->id;
    }

    /**
     * Returns array of cohorts that have to be synchrinized.
     * @return enrol_ubxsshib\ubxsshib_cohort[]
     */
    public static function find() {
        global $DB;

        $instances = $DB->get_records_sql("SELECT id, name, description, idnumber FROM {cohort} WHERE idnumber like 'ubgroup:%'");

        $cohorts = [];
        foreach ($instances as $instance) {
            array_push($cohorts, new ubxsshib_cohort($instance));
        }

        return $cohorts;
    }


    /**
     * Returns the group identifier.
     *
     * @return string groupid;
     */
    public function group() {
        preg_match_all($this->pattern, $this->instance->idnumber, $matches);
        return $matches[2][0];
    }

    /**
     * Returns sync mode.
     *
     * @return string mode
     */
    public function sync_mode() {
        preg_match_all($this->pattern, $this->instance->idnumber, $matches);
        return $matches[1][0];
    }

    /**
     * Return the underlying cohort instance.
     *
     * @return stdClass cohortinstance
     */
    public function cohort_instance() {
        return $this->instance;
    }

    /**
     * Indicates whether the supplied user is already member of this cohort.
     *
     * @global type $DB
     * @param stdClass $user
     * @return boolean isenroled
     */
    public function is_member($user) {
        return cohort_is_member($this->instance->id, $user->id);
    }

    /**
     * Returns users enroled thanks to this method.
     *
     * @return stdClass[] $users;
     */
    protected function get_members() {
        global $DB;

        $sql = 'SELECT U.* FROM {user} U INNER JOIN {cohort_members} CM on U.id = CM.userid WHERE CM.cohortid = ?';
        $users = $DB->get_records_sql($sql, [$this->instance->id]);

        return $users;
    }

    /**
     * Enrol the provided user through this method.
     *
     * @param stdClass $user
     * @return void
     */
    public function add_member($user) {
        cohort_add_member($this->instance->id, $user->id);
    }

    /**
     * Remove the provided user from the cohort.
     *
     * @param stdClass $user
     */
    public function remove_member($user) {
        cohort_remove_member($this->instance->id, $user->id);
    }

    /**
     * Add multiple members.
     *
     * @param stdCLass[] $users
     */
    public function add_members($users) {
        foreach ($users as $user) {
            $this->add_member($user);
        }
    }

    /**
     * Unenrol multiple users.
     *
     * @param stdClass[] $users
     */
    public function remove_members($users) {
        foreach ($users as $user) {
            $this->remove_member($user);
        }
    }

    /**
     * Synchronize enroled users with provided users.
     * Unenrol unlisted users and enrol users in list but not yet enroled.
     *
     * @param stdClass[] $users
     */
    public function sync_members($users) {
        $enroled = $this->get_members();
        $syncmode = $this->sync_mode();
        $enroledkeys = [];
        foreach ($enroled as $user) {
            $enroledkeys[$user->username] = $user;
        };

        $newkeys = [];
        foreach ($users as $user) {
            $newkeys[$user->username] = $user;
        };

        if ($syncmode == self::SYNCMODE_SYNC) {
            $tobeunenroled = array_diff_key($enroledkeys, $newkeys);
            $this->remove_members(array_values($tobeunenroled));
        }

        $toenrol = array_diff_key($newkeys, $enroledkeys);
        $this->add_members(array_values($users));
    }
}
