<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

/**
 * Class to manage ubxsshib enrolment method instances.
 *
 * @package    enrol_ubxsshib
 * @copyright  2018 - MAPI Université de Bordeaux
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

namespace enrol_ubxsshib;

defined('MOODLE_INTERNAL') || die();

/**
 * Class to manage ubxsshib enrolment method instances.
 *
 * @package    enrol_ubxsshib
 * @copyright  2018 - MAPI Université de Bordeaux
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */
abstract class ubxsshib_enrol_method {
    protected $instance = null;
    protected $plugin = null;

    const CODE_ETAPE = 'ETP';
    const CODE_ELEMENTPEDAGO = 'ELMT';
    const CODE_EPREUVE = 'EXAM';
    const NO_GROUP = -1;


    public static function get(\stdClass $plugininstance) {
        $m = null;
        $type = $plugininstance->customchar1;

        if (isset($type) && '' !== $type) {
            $mclass = '\enrol_ubxsshib\ubxsshib_enrol_' . strtolower($plugininstance->customchar1) . '_method';

            try {
                $m = new $mclass();
                $m->plugin = enrol_get_plugin('ubxsshib');
                $m->instance = $plugininstance;
            } catch (\Exception $e) {
                $m = false;
            }
        }

        return $m;
    }

    public function get_id() {
        return $this->instance->id;
    }

    /**
     * Returns array of methods matching provided code and type.
     *
     * @param string $code
     * @param string $data
     * @param int $status
     * @return enrol_ubxsshib\ubxsshib_enrol_method[] methods
     */
    public static function find($code = null, $codetype = null, $status = ENROL_INSTANCE_ENABLED) {
        global $DB;

        $criteria = [];
        $criteria['enrol'] = 'ubxsshib';
        $criteria['status'] = $status;
        if (isset($codetype)) {
            $criteria['customchar1'] = trim($codetype);
        }
        if (isset($code)) {
            $criteria['customchar2'] = trim($code);
        }

        $instances = $DB->get_records('enrol', $criteria);
        $methods = [];
        foreach ($instances as $instance) {
            $method = self::get($instance);
            if (false !== $method) {
                array_push($methods, $method);
            }
        }

        return $methods;
    }

    /**
     * Returns array of methods the provided user is enroled with.
     *
     * @param stdClass $user
     * @param string $codetype
     * @param array $except
     * @return enrol_ubxsshib\ubxsshib_enrol_method[] methods
     */
    public static function find_for_user(\stdClass $user, $codetype, array $except = []) {
        global $DB;

        $methods = [];

        $sql = 'SELECT E.* FROM {user_enrolments} UE INNER JOIN {enrol} E ON UE.enrolid = E.id ' .
            'WHERE E.enrol = ? AND UE.userid = ? AND E.customchar1 = ?';
        foreach ($except as $exception) {
            $sql .= ' AND NOT (E.customchar1 = \'' . $exception['type'] . '\' AND E.customchar2 = \'' . $exception['code'] . '\')';
        }

        $results = $DB->get_records_sql($sql, ['ubxsshib', $user->id, $codetype]);
        foreach ($results as $result) {
            array_push($methods, self::get($result));
        }

        return $methods;
    }

    /**
     * Returns the code type.
     *
     * @return string codetype;
     */
    public function code_type() {
        return $this->instance->customchar1;
    }

    /**
     * Returns the Apogee code.
     *
     * @return string code value
     */
    public function code_value() {
        return $this->instance->customchar2;
    }

    /**
     * Returns the group id.
     * @return int groupid
     */
    public function groupid() {
        return $this->instance->customint1;
    }

    /**
     * Indicates whether the method is active (visible + time matches active period).
     *
     * @return boolean isactive
     */
    public function is_active() {
        $now = time();
        $enrolinstance = $this->instance;
        return $enrolinstance->status == ENROL_INSTANCE_ENABLED &&
            ($enrolinstance->enrolstartdate == 0 ||
                ($enrolinstance->enrolstartdate != 0 && $enrolinstance->enrolstartdate < $now)) &&
            ($enrolinstance->enrolenddate == 0 ||
                ($enrolinstance->enrolenddate != 0 && $enrolinstance->enrolenddate > $now));
    }

    /**
     * Return the underlying enrol instance.
     *
     * @return stdClass enrolinstance
     */
    public function enrol_instance() {
        return $this->instance;
    }

    /**
     * Indicates whether the supplied user is already enroled through this method.
     *
     * @global type $DB
     * @param stdClass $user
     * @return boolean isenroled
     */
    public function is_enroled($user) {
        global $DB;
        $isenroled = $DB->record_exists('user_enrolments', [
            'userid' => $user->id,
            'enrolid' => $this->instance->id,
            ]
        );
        return $isenroled;
    }

    /**
     * Returns users enroled thanks to this method.
     *
     * @return stdClass[] $users;
     */
    protected function get_enroled_users() {
        global $DB;

        $sql = 'SELECT U.* FROM {user} U INNER JOIN {user_enrolments} UE on U.id = UE.userid WHERE UE.enrolid = ?';
        $users = $DB->get_records_sql($sql, [$this->instance->id]);

        return $users;
    }

    /**
     * Enrol the provided user through this method.
     *
     * @param stdClass $user
     */
    public function enrol_user($user) {
        // If user not already enroled.
        if (!$this->is_enroled($user)) {
            // Enrol user with grades recovering.
            $this->plugin->enrol_user($this->instance, $user->id, $this->instance->roleid, 0, 0, ENROL_USER_ACTIVE, true);
        }
        if (isset($this->instance->customint1) && $this->instance->customint1 != self::NO_GROUP) {
            groups_add_member($this->instance->customint1, $user->id, 'enrol_ubxsshib', $this->instance->id);
        }
        /* else {
            // TODO: Else verify roleid hasn't changed ?
            // TODO: Else verify groupid hasn't changed ?
        }
         */
    }

    /**
     * Unenrol the provided user from the course.
     *
     * @param stdClass $user
     */
    public function unenrol_user($user) {
        $this->plugin->unenrol_user($this->instance, $user->id);
    }

    /**
     * Enrol multiple users.
     *
     * @param stdCLass[] $users
     */
    public function enrol_users($users) {
        foreach ($users as $user) {
            $this->enrol_user($user);
        }
    }

    /**
     * Unenrol multiple users.
     *
     * @param stdClass[] $users
     */
    public function unenrol_users($users) {
        foreach ($users as $user) {
            $this->unenrol_user($user);
        }
    }

    /**
     * Synchronize enroled users with provided users.
     * Unenrol unlisted users and enrol users in list but not yet enroled.
     *
     * @param stdClass[] $users
     */
    public function sync_users($users) {
        $enroled = $this->get_enroled_users();
        $enroledkeys = [];
        foreach ($enroled as $user) {
            $enroledkeys[$user->username] = $user;
        }

        $newkeys = [];
        foreach ($users as $user) {
            $newkeys[$user->username] = $user;
        }

        $tobeunenroled = array_diff_key($enroledkeys, $newkeys);
        foreach ($tobeunenroled as $username => $user) {
            $this->unenrol_user($user);
            unset($enroledkeys[$username]);
        }

        $toenrol = array_diff_key($newkeys, $enroledkeys);
        foreach ($toenrol as $username => $user) {
            $this->enrol_user($user);
        }
    }
}
