<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

/**
 * Enrol page for UBxS shibboleth enrol plugin.
 *
 * @package    enrol_ubxsshib
 * @copyright  2014 Christophe Michenaud
 * @author     Christophe Michenaud
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

define('NO_OUTPUT_BUFFERING', true);

require_once('../../config.php');
require_once("$CFG->dirroot/user/lib.php");
require_once("$CFG->dirroot/enrol/ubxsshib/locallib.php");

use enrol_ubxsshib\ubxsshib_enrol_method;
use enrol_ubxsshib\form\force_enrol_form;
use enrol_ubxsshib\task\force_enrolments;
use core\task\manager;

global $PAGE, $USER;

$methodid   = required_param('mid', PARAM_INT);
$courseid   = required_param('courseid', PARAM_INT);
$confirm   = optional_param('confirm', 0, PARAM_INT);
$ispost = ($_SERVER['REQUEST_METHOD'] === 'POST');

$enrol = $DB->get_record('enrol', ['id' => $methodid], '*', MUST_EXIST);
if (!$enrol || ($enrol && $enrol->courseid != $courseid)) {
    throw new moodle_exception('methodmismatch', 'enrol_ubxsshib', new moodle_url('/enrol/instances.php',
        ['id' => $courseid]));
    /*print_error('methodmismatch', 'enrol_ubxsshib', new moodle_url('/enrol/instances.php',
                    ['id' => $courseid]));*/
}

$course = $DB->get_record('course', ['id' => $courseid], '*', MUST_EXIST);
$codetype = $enrol->customchar1;
$code = $enrol->customchar2;
$context = context_course::instance($courseid);

require_login($course);
require_capability('enrol/ubxsshib:force', $context);

$PAGE->set_url($CFG->wwwroot . '/enrol/ubxsshib/force_enrol.php', ['enrolid' => $methodid]);
$PAGE->set_pagelayout('admin');
$PAGE->set_title($enrol->name);
$PAGE->set_heading($course->fullname);
navigation_node::override_active_url(new moodle_url('/enrol/instances.php', ['id' => $courseid]));

echo $OUTPUT->header();
echo $OUTPUT->heading($enrol->name);

if (!$confirm) {
    echo $OUTPUT->confirm(get_string('sync_warning', 'enrol_ubxsshib', ['methodname' => $enrol->name, 'methodcode' => $code]),
        new single_button(new moodle_url('/enrol/ubxsshib/force_enrol.php',
            ['mid' => $methodid, 'courseid' => $courseid, 'confirm' => 1, 'sesskey' => sesskey()]),
            get_string('sync_confirm', 'enrol_ubxsshib')),
        new single_button(new moodle_url('/enrol/instances.php',
            ['id' => $courseid]),
            get_string('cancel'), false));
    echo $OUTPUT->footer();
    exit;
}

require_sesskey();
\core\session\manager::write_close();


if ($enrol != null) {
    $mgr = new \enrol_ubxsshib\ldap_provider();
    $grp = $mgr->get_group($codetype, $code, false);
    if ($grp != null) {
        echo("<strong>$grp->code - $grp->description</strong><br /><br />");

        if ($ispost) {
            $method = ubxsshib_enrol_method::get($enrol);
            /** @var \enrol_ubxsshib\task\force_enrolments $task */
            $task = new force_enrolments();
            $task->set_blocking(false);
            $task->set_custom_data([
                    'methodid' => $method->get_id(),
                    'context' => $context,
                    'userid' => $USER->id,
                    'courseid' => $course->id,
                ]
            );
            $task->set_userid($USER->id);
            $queued = manager::queue_adhoc_task($task, true);
            $message = $queued ?
                get_string('forceenrol_confirm_message_ok', 'enrol_ubxsshib') :
                get_string('forceenrol_confirm_message_pending', 'enrol_ubxsshib');
            echo($message);
        }
    } else {
        $message = get_string('forceenrol_group_notfound', 'enrol_ubxsshib');
        echo($message);
    }
    unset($mgr);

    echo '<br />' . $OUTPUT->single_button(new moodle_url('/enrol/instances.php',
            ['id' => $courseid]),
            get_string('back'));
}
echo $OUTPUT->footer();
