<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

/**
 * UBxS shibboleth enrolment plugin.
 *
 * @package    enrol_ubxsshib
 * @copyright  2011 Christophe Michenaud
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

defined('MOODLE_INTERNAL') || die();

use enrol_ubxsshib\user_info;
use enrol_ubxsshib\ldap_provider;
use enrol_ubxsshib\ubxsshib_enrol_method;
use enrol_ubxsshib\ubxsshib_cohort;

require_once($CFG->dirroot.'/enrol/ubxsshib/locallib.php');

class enrol_ubxsshib_plugin extends enrol_plugin {
    public function get_description_text($instance) {
        return "Enrol users according to Apogee code";
    }

    public function roles_protected() {
        return true;
    }

    public function allow_enrol(stdClass $instance) {
        return true;
    }

    public function allow_unenrol_user(stdClass $instance, stdClass $ue) {
        return true;
    }

    public function allow_unenrol(stdClass $instance) {
        return true;
    }

    public function allow_manage(stdClass $instance) {
        return true;
    }

    public function show_enrolme_link(stdClass $instance) {
        return false;
    }

    public function try_autoenrol(stdClass $instance) {
        global $USER;

        return false;
    }

    public function try_guestaccess(stdClass $instance) {
        global $USER;

        return false;
    }

    public function sync_user_enrolments($user) {
        global $CFG, $DB;

        $inhibitall = $this->get_config('inhibit_all', false);
        if ($inhibitall) {
            debugging("ubxsshib enrolment methods are disabled", DEBUG_DEVELOPER);
            return false;
        }

        if (($user->auth == 'cas') || ($user->auth == 'shibboleth' && endswith($user->username, '@u-bordeaux.fr'))) {
            $repository = new ldap_provider();
            /** @var user_info $userinfos */
            $userinfos = $repository->get_userinfo($user->username, true);

            // Setting system course creator status if authorized.
            if ($userinfos->mustbesystemcoursecreator) {
                enrol_ubxsshib_set_system_coursecreation_rights($user);
            }

            // Setting coursecreator status to authorized users on sandbox course category.
            if ($userinfos->mustbesandboxcoursecreator) {
                enrol_ubxsshib_set_sandbox_coursecreation_rights($user);
            }

            // Retrieving Apogee codes for current user.
            $supannetapes = $userinfos->etapes;
            $supannelempedago = $userinfos->elements;
            $supannepreuves = $userinfos->epreuves;

            // Should we add additional codes set up in config ?
            $usertooverride = $this->get_config('usertooverride');
            $etpforcedattributes = $this->get_config('etp_forcedattributes');
            $elpforcedattributes = $this->get_config('elp_forcedattributes');
            $examforcedattributes = $this->get_config('exam_forcedattributes');
            if ($usertooverride != '' &&
                ($etpforcedattributes != '' || $elpforcedattributes != ''|| $examforcedattributes != '')) {
                if (isset($user) && $user->username == $usertooverride) {
                    $etps = explode(';', str_replace(' ', '', $etpforcedattributes));
                    $elps = explode(';', str_replace(' ', '', $elpforcedattributes));
                    $exams = explode(';', str_replace(' ', '', $examforcedattributes));
                    $userinfos->etapes = array_merge($supannetapes, $etps);
                    $userinfos->elements = array_merge($supannelempedago, $elps);
                    $userinfos->epreuves = array_merge($supannepreuves, $exams);
                }
            }

            // Logging Apogee attributes provided by Shibboleth during login phase.
            $logattributes = $this->get_config('logattributes');
            if ($logattributes) {
                enrol_ubxsshib_log_apogee_attributes($userinfos);
            }

            // Syncronizing user enrolments based on Apogee.
            $codes = [
                ubxsshib_enrol_method::CODE_ETAPE => $userinfos->etapes,
                ubxsshib_enrol_method::CODE_ELEMENTPEDAGO => $userinfos->elements,
                ubxsshib_enrol_method::CODE_EPREUVE => $userinfos->epreuves,
            ];

            enrol_ubxsshib_sync_user_enrolments($user, $codes);
        }

        return true;
    }

    /**
     * Forces synchronisation of all enrolments with Apogee.
     *
     * @param progress_trace $trace
     * @param int|null $onemethod limit sync to one method, null if all Apogee methods
     * @return void
     */
    public function sync_enrolments(progress_trace $trace, $onemethod = null) {
        global $CFG, $DB;

        // A little bit more memory.
        core_php_time_limit::raise();
        raise_memory_limit(MEMORY_EXTRA);

        if ($onemethod) {
            $methods = [$onemethod];
        } else {
            $methods = ubxsshib_enrol_method::find(null, null, ENROL_INSTANCE_ENABLED);
        }
        $activemethods = array_filter($methods, function(ubxsshib_enrol_method $m) {
            return $m->is_active();
        });
        $trace->output('Active methods to synchronize : ' . count($activemethods));

        $methodslist = [];
        foreach ($activemethods as $method) {
            $methodkey = $method->code_type() . '.' . $method->code_value();
            if (!array_key_exists($methodkey, $methodslist)) {
                $methodslist[$methodkey] = [];
            }
            $mlist = $methodslist[$methodkey];
            array_push($mlist, $method);
            $methodslist[$methodkey] = $mlist;
        }
        $trace->output('Users groups involved : ' . count(array_keys($methodslist)));

        $isdebugactive = (!empty($CFG->debug) && ($CFG->debug == true));
        $grpprovider = new ldap_provider();
        foreach ($methodslist as $key => $list) {
            $codes = explode('.', $key);
            $codetype = $codes[0];
            $code = $codes[1];
            $trace->output("Processing Group $codetype-$code...");
            $grp = $grpprovider->get_group($codetype, $code, true);
            if (isset($grp)) {
                $trace->output("    ... group found");
                $users = enrol_ubxsshib_get_users_from_group($grp, true);
                if (count($users) == 0) {
                    $trace->output("    ... no member in LDAP group => no synchronization");
                } else {
                    $trace->output("    ... " . count($grp->members) . " member(s) found in LDAP group");
                    $trace->output("    ... " . count($list) . " method(s) found for this group");
                    /** @var ubxsshib_enrol_method $method */
                    foreach ($list as $method) {
                        if ($isdebugactive) {
                            $trace->output("    ... processing method #" . $method->get_id() . " / course #" . $method->enrol_instance()->courseid);
                        }
                        try {
                            $method->sync_users($users);
                        } catch (\Throwable $e) {
                            $trace->output("    ... exception raised : " . $e->getMessage());
                        }
                    }
                }
                unset($users);
            } else {
                $trace->output('    ...  not found => no synchronization');
            }
        }
    }

    /**
     * Forces synchronisation of all cohorts with Apogee.
     *
     * @param progress_trace $trace
     * @param int|null $onecohort limit sync to one cohort, null if all cohorts
     * @return void
     */
    public function sync_cohorts(progress_trace $trace, $onecohort = null) {
        global $CFG, $DB;

        // A little bit more memory.
        core_php_time_limit::raise();
        raise_memory_limit(MEMORY_EXTRA);

        if ($onecohort) {
            $cohorts = [$onecohort];
        } else {
            $cohorts = ubxsshib_cohort::find();
        }
        $trace->output('Cohorts to synchronize : ' . count($cohorts));

        $cohortslist = [];
        foreach ($cohorts as $cohort) {
            $cohortkey = $cohort->group();
            if (!array_key_exists($cohortkey, $cohortslist)) {
                $cohortslist[$cohortkey] = [];
            }
            $clist = $cohortslist[$cohortkey];
            array_push($clist, $cohort);
            $cohortslist[$cohortkey] = $clist;
        }
        $trace->output('Users groups involved : ' . count(array_keys($cohortslist)));

        $grpprovider = new ldap_provider();
        foreach ($cohortslist as $key => $list) {
            $code = $key;
            $trace->output("Processing group $code");
            $members = $grpprovider->search_users(['memberOf' => 'cn='.$code]);
            $keyedmembers = [];
            foreach ($members as $member) {
                $keyedmembers[$member->username] = $member;
            }

            $memberscount = count(array_keys($keyedmembers));
            if ($memberscount == 0) {
                $trace->output("    ... no members in group : no synchronization.");
            } else {
                $trace->output("    ..." . $memberscount . " member(s) found in group");
                $users = enrol_ubxsshib_get_users($keyedmembers, true);
                $trace->output("    ... " . count($list) . " cohort(s) for this group");
                /** @var ubxsshib_cohort $cohort */
                foreach ($list as $cohort) {
                    $cohort->sync_members($users);
                }
            }
        }
        unset($grpprovider);

        $trace->output('All cohorts synchronized');
        $trace->finished();
    }

    /**
     * Restore instance and map settings.
     *
     * @param restore_enrolments_structure_step $step
     * @param stdClass $data
     * @param stdClass $course
     * @param int $oldid
     */
    public function restore_instance(restore_enrolments_structure_step $step, stdClass $data, $course, $oldid) {
        // Changing group id reference.
        $oldgroupid = $data->customint1;
        if (!empty($oldgroupid)) {
            $data->customint1 = $step->get_mappingid('group', $oldgroupid);
        }

        $instanceid = $this->add_instance($course, (array)$data);
        $step->set_mapping('enrol', $oldid, $instanceid);
    }

    /**
     * Restore user enrolment.
     *
     * @param restore_enrolments_structure_step $step
     * @param stdClass $data
     * @param stdClass $instance
     * @param int $oldinstancestatus
     * @param int $userid
     */
    public function restore_user_enrolment(restore_enrolments_structure_step $step, $data, $instance, $userid, $oldinstancestatus) {
        global $DB;

        $this->enrol_user($instance, $userid, null, $data->timestart, $data->timeend, $data->status, true);
    }

    /**
     * Restore role assignment.
     *
     * @param stdClass $instance
     * @param int $roleid
     * @param int $userid
     * @param int $contextid
     */
    public function restore_role_assignment($instance, $roleid, $userid, $contextid) {
        role_assign($roleid, $userid, $contextid, '', 0);
    }

    /**
     * Restore user group membership.
     * @param stdClass $instance
     * @param int $groupid
     * @param int $userid
     */
    public function restore_group_member($instance, $groupid, $userid) {
        global $CFG;
        require_once("$CFG->dirroot/group/lib.php");

        groups_add_member($groupid, $userid, 'enrol_ubxsshib', $instance->id);
    }

    /**
     * Returns defaults for new instances.
     * @since Moodle 3.1
     * @return array
     */
    public function get_instance_defaults() {
        global $DB;

        $studentrole = $DB->get_record('role', ['shortname' => 'student'], 'id', MUST_EXIST);

        $now = (new \DateTime())->getTimestamp();
        $month = (int)date("n", $now);
        $year = (int)date("Y", $now);
        if ($month > 6) {
            $year += 1;
        }
        $enddate = strtotime("1st july $year");

        return [
            'roleid' => $studentrole->id,
            'enrolstartdate' => 0,
            'enrolenddate' => $enddate,
        ];
    }

    /**
     * Returns enrolment instance manage link.
     *
     * By defaults looks for manage.php file and tests for manage capability.
     *
     * @param navigation_node $instancesnode
     * @param stdClass $instance
     * @return moodle_url;
     */
    public function add_course_navigation($instancesnode, stdClass $instance) {
        if ($instance->enrol !== 'ubxsshib') {
            throw new coding_exception('Invalid enrol instance type!');
        }

        $context = context_course::instance($instance->courseid);
        if (has_capability('enrol/ubxsshib:config', $context)) {
            $managelink = new moodle_url('/enrol/ubxsshib/edit.php',
                ['courseid' => $instance->courseid, 'id' => $instance->id]);
            $instancesnode->add($this->get_instance_name($instance),
                $managelink, navigation_node::TYPE_SETTING);
        }
    }

    /**
     * Returns edit icons for the page with list of instances
     * @param stdClass $instance
     * @return array
     */
    public function get_action_icons(stdClass $instance) {
        global $OUTPUT;

        if ($instance->enrol !== 'ubxsshib') {
            throw new coding_exception('invalid enrol instance!');
        }
        $context = context_course::instance($instance->courseid);

        $icons = [];

        if (has_capability('enrol/ubxsshib:force', $context)) {
            $forceenrollink = new moodle_url("/enrol/ubxsshib/force_enrol.php", ['courseid' => $instance->courseid, 'mid' => $instance->id]);
            $icons[] = $OUTPUT->action_icon($forceenrollink, new pix_icon('t/enrolusers', get_string('forceenrol_button', 'enrol_ubxsshib'), 'core',
                ['class' => 'iconsmall']));
        }

        if (has_capability('enrol/ubxsshib:manage', $context)) {
            $editlink = new moodle_url("/enrol/ubxsshib/edit.php", ['courseid' => $instance->courseid, 'id' => $instance->id]);
            $icons[] = $OUTPUT->action_icon($editlink, new pix_icon('t/edit', get_string('edit'), 'core',
                ['class' => 'iconsmall']));
        }

        return $icons;
    }

    /**
     * Backup execution step hook to annotate custom fields.
     *
     * @param backup_enrolments_execution_step $step
     * @param stdClass $enrol
     */
    public function backup_annotate_custom_fields(backup_enrolments_execution_step $step, stdClass $enrol) {
        // Override as necessary to annotate custom fields in the enrol table.
    }

    /**
     * Returns link to page which may be used to add new instance of enrolment plugin in course.
     * @param int $courseid
     * @return moodle_url page url
     */
    public function get_newinstance_link($courseid) {
        $context = context_course::instance($courseid, MUST_EXIST);

        if (!has_capability('moodle/course:enrolconfig', $context) || !has_capability('enrol/ubxsshib:config', $context)) {
            return null;
        }

        // Multiple instances supported.
        return new moodle_url('/enrol/ubxsshib/edit.php', ['courseid' => $courseid]);
    }

    public function can_delete_instance($instance) {
        $context = context_course::instance($instance->courseid);
        return has_capability('enrol/ubxsshib:config', $context);
    }

    public function can_hide_show_instance($instance) {
        $context = context_course::instance($instance->courseid);
        return has_capability('enrol/ubxsshib:config', $context);
    }
}
