<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

/**
 * Local stuff for ubxsshib enrolment plugin.
 *
 * @package    enrol_ubxsshib
 * @copyright  2011 Christophe Michenaud
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

defined('MOODLE_INTERNAL') || die();

use enrol_ubxsshib\ubxsshib_enrol_method;

require_once($CFG->dirroot . '/user/lib.php');
require_once($CFG->libdir . '/grouplib.php');
require_once($CFG->dirroot . '/group/lib.php');

/**
 * Writes extracted attributes to Moodle log, if set in plugin config page.
 *
 * @param \enrol_ubxsshib\user_info $userinfo
 * @return void
 */
function enrol_ubxsshib_log_apogee_attributes(\enrol_ubxsshib\user_info $userinfo) {
    global $DB;

    $rec = new stdClass();
    $rec->username = $userinfo->username;
    $rec->logintime = time();
    $etapes = $userinfo->etapes;
    if (isset($etapes) && count($etapes) > 0) {
        $rec->etuetape = implode(';', $etapes);
    } else {
        $rec->etuetape = '-none-';
    }

    $elements = $userinfo->elements;
    if (isset($elements) && count($elements) > 0) {
        $rec->etuelempedago = implode(';', $elements);
    } else {
        $rec->etuelempedago = '-none-';
    }

    $epreuves = $userinfo->epreuves;
    if (isset($epreuves) && count($epreuves) > 0) {
        $rec->etuepreuve = implode(';', $epreuves);
    } else {
        $rec->etuepreuve = '-none-';
    }

    $DB->insert_record('ubxsshib_login_log', $rec, false);
}

/**
 * Indicates if extracted code values for etape are valid or not.
 *
 * @param array $etapes List of code values for etapes
 * @return bool
 */
function enrol_ubxsshib_arevalid_etp($etapes) {
    $errorvalues = ['DB_ERROR', 'supannEtuEtapeDefault'];
    $iserror = false;
    foreach ($etapes as $etape) {
        if (in_array($etape, $errorvalues, false)) {
            $iserror = true;
            break;
        }
    }
    return !$iserror;
}

/**
 * Indicates if extracted code values for elements are valid or not.
 *
 * @param array $elements List of code values for elements
 * @return bool
 */
function enrol_ubxsshib_arevalid_elp($elements) {
    $errorvalues = ['DB_ERROR', 'supannEtuElementPedagogiqueDefault'];
    $iserror = false;
    foreach ($elements as $element) {
        if (in_array($element, $errorvalues, false)) {
            $iserror = true;
            break;
        }
    }
    return !$iserror;
}

/**
 * Indicates if extracted code values for epreuve are valid or not.
 *
 * @param array $elements List of code values for elements
 * @return bool
 */
function enrol_ubxsshib_arevalid_exam($elements) {
    $errorvalues = ['DB_ERROR', 'supannEtuExamDefault'];
    $iserror = false;
    foreach ($elements as $element) {
        if (in_array($element, $errorvalues, false)) {
            $iserror = true;
            break;
        }
    }
    return !$iserror;
}

/**
 * Assigns user course creator role at system level.
 *
 * @param stdClass $user User to be assign course creator role.
 * @return void
 */
function enrol_ubxsshib_set_system_coursecreation_rights($user) {
    $sitecontext = context_system::instance();

    if ($user != null && $roles = explode(',', get_config('enrol_ubxsshib', 'systemroleids'))) {
        // Getting selected roles in plugin settings.
        while (count($roles) > 0) {
            $roleid = array_shift($roles);
            role_assign(intval($roleid), $user->id, $sitecontext->id);
        }
    }
}

/**
 * Assigns user course creator role into the sandbox.
 *
 * @param stdClass $user User to be assign course creator role.
 * @return void
 */
function enrol_ubxsshib_set_sandbox_coursecreation_rights($user) {
    global $DB;

    // Getting sandbox category id.
    $sandboxid = get_config('enrol_ubxsshib', 'sandboxid');

    if (isset($sandboxid)) {
        // Getting course creator role.
        $coursecreatorrole = $DB->get_record('role', ['shortname' => 'coursecreator']);

        // Getting Context of training category (Bac a sable).
        $sandboxcontext = context_coursecat::instance($sandboxid);

        // Assigning coursecreator role to user in training course category.
        if ($sandboxcontext != false && $coursecreatorrole != null) {
            role_assign($coursecreatorrole->id, $user->id, $sandboxcontext->id);
        }
    }
}

/**
 * Enrols/unenrols the user regarding his Apogee codes.
 *
 * @param stdClass $user
 */
function enrol_ubxsshib_sync_user_enrolments($user, $codes) {
    foreach ($codes as $codetype => $codevalues) {

        // Unenrol user from active methods of this type.
        $current = ubxsshib_enrol_method::find_for_user($user, $codetype);
        $tounenrol = array_filter($current, function($m, $k) use($codevalues) {
            return ($m->is_active() && !in_array($m->code_value(), $codevalues));

        }, ARRAY_FILTER_USE_BOTH);
        foreach ($tounenrol as $method) {
            $method->unenrol_user($user);
        }

        // Enrol user from active methods of this type.
        foreach ($codevalues as $value) {
            $toenrol = ubxsshib_enrol_method::find($value, $codetype);
            foreach ($toenrol as $method) {
                if ($method->is_active()) {
                    $method->enrol_user($user);
                }
            }
        }
    }
}

require_once($CFG->libdir . '/adminlib.php');

function endswith($mainstr, $pattern) {
    $mainlen = strlen($mainstr);
    $patternlen = strlen($pattern);
    if ($patternlen > $mainlen) {
        return false;
    }
    return substr_compare($mainstr, $pattern, $mainlen - $patternlen, $patternlen) === 0;
}

function enrol_ubxsshib_get_users($groupmembers, $forcecreate = true) {
    global $DB, $CFG;

    $plugin = enrol_get_plugin('ubxsshib');
    $authmethod = $plugin->get_config('auth_method');
    $cache = \cache::make('enrol_ubxsshib', 'userinfo');

    // Finding members without DB id.
    $missingids = array_filter(
        $groupmembers,
        function($v, $k) {
            return $v->id == 0;
        },
        ARRAY_FILTER_USE_BOTH);
    if (count($missingids) > 0) {
        // Getting usernames to lookup in DB.
        $missingnames = [];
        foreach ($missingids as $member) {
            $missingnames[] = $member->username;
        }
        list($sqlnames, $paramnames) = $DB->get_in_or_equal($missingnames);
        $sql = "SELECT * from {user} WHERE username $sqlnames";
        $users = $DB->get_records_sql($sql, $paramnames);
        // Setting id attribute to cached userinfo.
        foreach ($users as $user) {
            $member = $cache->get($user->username);
            $member->id = $user->id;
            $cache->set($user->username, $member);
            $groupmembers[$user->username] = $member;
        }
    }
    unset($missingids);

    if ($forcecreate) {
        // Processiong members unknown in DB.
        $unknownids = array_filter(
            $groupmembers,
            function($v, $k) {
                return $v->id == 0;
            },
            ARRAY_FILTER_USE_BOTH);
        foreach ($unknownids as $member) {
            $member->auth = $authmethod;
            $member->mnethostid = $CFG->mnet_localhost_id;
            $member->lang = 'fr';
            $member->confirmed = 1;
            $newid = user_create_user($member, false, true);
            $member->id = $newid;
            $userid = $newid;
            $cache->set($member->username, $member);
            $groupmembers[$member->username] = $member;
        }
        unset($unknownids);
    }

    // Getting all valid members (ie. with previous id, updated with id, newly created).
    $validmembers = array_filter(
        $groupmembers,
        function($v, $k) {
            return $v->id > 0;
        },
        ARRAY_FILTER_USE_BOTH);
    $validids = [];
    if (count($validmembers) > 0) {
        // Getting all ids to exec. only one DB query.
        foreach ($validmembers as $member) {
            $validids[] = $member->id;
        }
        list($sqlids, $paramids) = $DB->get_in_or_equal($validids);
        $sql = "SELECT * from {user} WHERE id $sqlids";
        $users = $DB->get_records_sql($sql, $paramids);
    } else {
        $users = [];
    }
    unset($validmembers);

    return $users;
}
function enrol_ubxsshib_get_users_from_group($group, $forcecreate = true) {
    if ($group == null) {
        return [];
    }

    return enrol_ubxsshib_get_users($group->members, $forcecreate);
}

function enrol_ubxsshib_check_coursegroup_exists($groupid) {
    return groups_group_exists($groupid);
}

function enrol_ubxsshib_remove_group_from_method($groupid) {
    global $DB;

    if (is_numeric($groupid)) {
        $DB->set_field('enrol', 'customint1', null, ['enrol' => 'ubxsshib', 'customint1' => $groupid]);
    }
}
